@@ -13,6 +13,7 @@ import subprocess
 import sys
 import threading
 import time
+import traceback
 
 import OpenSSL.SSL
 import pytest
@@ -40,7 +41,7 @@ IS_WIN2016 = (
     IS_WINDOWS
     # pylint: disable=unsupported-membership-test
     and b'Microsoft Windows Server 2016 Datacenter' in subprocess.check_output(
-        ('systeminfo', ),
+        ('systeminfo',),
     )
 )
 IS_LIBRESSL_BACKEND = ssl.OPENSSL_VERSION.startswith('LibreSSL')
@@ -149,7 +150,7 @@ def tls_ca_certificate_pem_path(ca):
 def tls_certificate(ca):
     """Provide a leaf certificate via fixture."""
     interface, host, port = _get_conn_data(ANY_INTERFACE_IPV4)
-    return ca.issue_server_cert(ntou(interface), )
+    return ca.issue_server_cert(ntou(interface))
 
 
 @pytest.fixture
@@ -166,10 +167,46 @@ def tls_certificate_private_key_pem_path(tls_certificate):
         yield cert_key_pem
 
 
+def _thread_except_hook(exceptions, args):
+    """Append uncaught exception ``args`` in threads to ``exceptions``."""
+    if issubclass(args.exc_type, SystemExit):
+        return
+    # cannot store the exception, it references the thread's stack
+    exceptions.append((
+        args.exc_type,
+        str(args.exc_value),
+        ''.join(
+            traceback.format_exception(
+                args.exc_type, args.exc_value, args.exc_traceback,
+            ),
+        ),
+    ))
+
+
+@pytest.fixture
+def thread_exceptions():
+    """Provide a list of uncaught exceptions from threads via a fixture.
+
+    Only catches exceptions on Python 3.8+.
+    The list contains: ``(type, str(value), str(traceback))``
+    """
+    exceptions = []
+    # Python 3.8+
+    orig_hook = getattr(threading, 'excepthook', None)
+    if orig_hook is not None:
+        threading.excepthook = functools.partial(
+            _thread_except_hook, exceptions,
+        )
+    try:
+        yield exceptions
+    finally:
+        if orig_hook is not None:
+            threading.excepthook = orig_hook
+
+
 @pytest.mark.parametrize(
     'adapter_type',
     (
-        'builtin',
         'pyopenssl',
     ),
 )
@@ -212,6 +249,7 @@ def test_ssl_adapters(
 @pytest.mark.parametrize(
     'adapter_type',
     (
+        'builtin',
         'pyopenssl',
     ),
 )
@@ -358,7 +396,7 @@ def test_tls_client_auth(
         )
         if not six.PY2:
             if IS_MACOS and IS_PYPY and adapter_type == 'pyopenssl':
-                expected_substrings = ('tlsv1 alert unknown ca', )
+                expected_substrings = ('tlsv1 alert unknown ca',)
         if (
                 tls_verify_mode in (
                     ssl.CERT_REQUIRED,
@@ -394,7 +432,10 @@ def test_tls_client_auth(
             ) if (
                 IS_GITHUB_ACTIONS_WORKFLOW
                 and IS_LINUX
-            ) else ()
+            ) else (
+                "('Connection aborted.', "
+                "BrokenPipeError(32, 'Broken pipe'))",
+            )
         assert any(e in err_text for e in expected_substrings)
 
 
@@ -416,6 +457,7 @@ def test_tls_client_auth(
     ),
 )
 def test_ssl_env(
+        thread_exceptions,
         recwarn,
         mocker,
         tls_http_server, adapter_type,
@@ -432,7 +474,7 @@ def test_ssl_env(
         'idna.core.ulabel',
         return_value=ntob('127.0.0.1'),
     ):
-        client_cert = ca.issue_cert(ntou('127.0.0.1'),)
+        client_cert = ca.issue_cert(ntou('127.0.0.1'))
 
     with client_cert.private_key_and_cert_chain_pem.tempfile() as cl_pem:
         tls_adapter_cls = get_ssl_adapter_class(name=adapter_type)
@@ -524,6 +566,15 @@ def test_ssl_env(
             )
         pytest.fail(msg)
 
+    # to perform the ssl handshake over that loopback socket,
+    # the builtin ssl environment generation uses a thread
+    for _, _, trace in thread_exceptions:
+        print(trace, file=sys.stderr)
+    assert not thread_exceptions, ': '.join((
+        thread_exceptions[0][0].__name__,
+        thread_exceptions[0][1],
+    ))
+
 
 @pytest.mark.parametrize(
     'ip_addr',
@@ -611,8 +662,10 @@ def test_http_over_https_error(
         fqdn = '[{}]'.format(fqdn)
 
     expect_fallback_response_over_plain_http = (
-        (adapter_type == 'pyopenssl'
-         and (IS_ABOVE_OPENSSL10 or not six.PY2))
+        (
+            adapter_type == 'pyopenssl'
+            and (IS_ABOVE_OPENSSL10 or not six.PY2)
+        )
         or PY27
     ) or (
         IS_GITHUB_ACTIONS_WORKFLOW
